import { z } from "zod";
/**
 * `HeaderValue` type represents what can be used to add HTTP header values. This
 * type follows the `headers` type required by the fetch API itself.
 *
 * Note that this is NOT A JSON stringifiable type, since headers for the fetch API
 * do not accept arbitrary objects.
 *
 * Exporting this type so that you can explicitly type your Header objects
 * with this to ensure that it is correctly typed at point of value definition
 * instead of only type checking when you call the `.header` method.
 *
 * Note that although traditionally primitive values like numbers and boolean were
 * allowed, it is only because they were implicitly converted to strings before they
 * were sent out, as the HTTP spec requires header values to be strings. Thus this
 * stricter type makes it more obvious and enforces it at the user level
 *
 * https://stackoverflow.com/questions/34152142/possible-types-of-a-http-header-value
 */
type HeaderValue = Exclude<HeadersInit, Headers>;
/**
 * ## `Header` Type
 * Header can either be,
 * 1. A `HeaderValue`
 * 2. A function that returns a `HeaderValue` or undefined
 * 3. A function that returns a Promise that resolves to a `HeaderValue` or undefined
 *
 * Header cannot be `undefined` directly because the `header` method cannot be directly called
 * with `undefined` because that wouldn't make any sense.
 *
 * Exporting this type so that you can explicitly type your Header values
 * with this to ensure that it is correctly typed at point of value definition
 * instead of only type checking when you call the `.header` method.
 *
 * ## `Header` function types
 * The function types can return undefined instead of a `HeaderValue` because the API call should
 * still run even if the header cannot be generated. For e.g. if using a function that reads auth
 * token from local storage and only return a `HeaderValue` with the token if it exists. This will
 * not cause any errors because spreading `undefined` into an object will just do nothing
 * `{ something: true, ...undefined }` will just be `{ something: true }`
 *
 * If function throws, the API call will be cancelled and the error will be caught by the 'safe'
 * 'run' methods before they are bubbled up and returned to the library caller.
 *
 * ## Function return types
 * the return type can be undefined since spreading it like ...undefined will just do nothing
 * this is because if my header generation function say to get a JWT if it gets nothing, it can just
 * dont return, and let it be undefined
 */
type Header = HeaderValue | (() => HeaderValue | undefined) | (() => Promise<HeaderValue | undefined>);
/**
 * All the supported HTTP methods.
 * Ref: https://fetch.spec.whatwg.org/#methods
 *
 * Based on the reference link, although there are other HTTP methods, they are not
 * supported by the `fetch` spec, and therefore not included in this union type as
 * even if you could pass it to the `fetch` function, it will just fail.
 *
 * `HEAD` and `OPTIONS` HTTP methods are quite special and low level and extremely
 * rarely used, therefore these are the only 2 methods which do not have wrapper
 * methods to easily create `Fetch` instances from `Builder` instance. However this
 * can be solved by library users simply by using the `HTTP` method on `Builder`.
 *
 * Exporting this type so that you can explicitly type your HTTP method strings
 * as needed to ensure that they are correctly typed at point of value definition
 * instead of only type checking when you call the constructor.
 */
type HTTPMethod = "GET" | "POST" | "PUT" | "PATCH" | "DELETE" | "HEAD" | "OPTIONS";
/**
 * `JsonResponse` is an alias for the `any` type because it follows the
 * return type of `ResponseBody.json()` which returns `any`. So there is
 * no point in making this a stricter type if the type is going to be
 * widened later on in the actual parsing since that would override this.
 */
type JsonResponse = any;
/**
 * Unifying all possible errors that can be thrown by fetch into
 * a single error type, covering possible errors thrown by header
 * function(s) too with the generic `Error` class.
 *
 * See this reference on all errors that can be thrown:
 * https://developer.mozilla.org/en-US/docs/Web/API/fetch#exceptions
 *
 * RequestError is exported from the library so that users can inspect this union type.
 */
type RequestError = 
// AbortError
DOMException | TypeError | Error;
/**
 * Expected data type returned from all `Fetch` run methods that do value extraction.
 *
 * Use satisfies to typecheck, to ensure that the return type will not be generalized
 * into the opaque `ApiResponse<T>` where users would not be able to see the shape of
 * the returned data clearly.
 */
type ApiResponse<T> = {
    ok: Response["ok"];
    status: Response["status"];
    headers: Response["headers"];
    data: T;
};
/**
 * Type of a generic type predicate, used for response validation.
 */
type Validator<T> = (data: unknown) => data is T;
/**
 * Class used to create an Object Oriented `Fetch` abstraction
 *
 * This object oriented approach gives users a easy to use chainable interface to build their API calls
 */
declare class Fetch {
    #private;
    readonly method: HTTPMethod;
    /**
     * Low level constructor API that should not be used directly by library users.
     * This is only used by the `Builder` class to construct a new instance after
     * using the builder pattern.
     *
     * ### Important
     * When passing in the default options and headers, make sure to pass in a new
     * object and array, as these will be passed in by reference, which means that
     * the original options object and header array will be mutated when `Fetch`
     * instance methods mutate the values on the instance itself.
     */
    constructor(method: HTTPMethod, url: string, defaultOpts: RequestInit, headers: Array<Header>);
    /**
     * Method to set `RequestInit` object for this one API call.
     *
     * ### When to use this method?
     * Use this to set custom RequestInit parameters for this specific `oof` instance's
     * fetch method call. See below on when should you use this method.
     *
     * Note that the options set using this method cannot override the headers set with
     * the `headers` method, cannot override the HTTP `method` set in the constructor
     * and cannot override the body value set by any of the 'body' methods.
     *
     * Which is why:
     * - If you want to set request body, use the `.body` or `.bodyJSON` method instead.
     * - If you need to set a request header, use the `.header` method instead.
     * - If you want to set HTTP request method, use one of the static methods on `Builder`
     *
     * Do not use this unless you have a specific option to pass in e.g. cache: "no-cache"
     * for this one specific API call only and nothing else. If there are options that you
     * want to set for all API request calls, use `Builder`'s `.setDefaultOptions` method
     * instead. Note that any options value set using this method will also override the
     * default options set.
     *
     * This method merges the provided options with the previously set options, so if any
     * default options can be overwritten there. Note that this is a shallow merge and not
     * a deepmerge.
     *
     * @returns Returns the current instance to let you chain method calls
     */
    options(opts: RequestInit): Fetch;
    /**
     * Add Header(s) to include in the API call.
     *
     * Accepts plain header objects, functions and async functions.
     *
     * Functions passed in will be called right before the API call to generate a header object,
     * to delay generating certain header values like a time limited auth token or recaptcha.
     *
     * This method can be called multiple times, and all the header objects will be combined.
     * If there are duplicate headers, the latter will be used.
     *
     * If you prefer, this method is also a variadic function that accepts multiple arguments
     * of the same type so that you can call this method just once if you have all the headers
     * instead of invoking this method multiple times.
     *
     * The `headers` function parameter forces the caller to pass in at least one argument
     * for this variadic method. See https://stackoverflow.com/a/72286990
     *
     * @returns Returns the current instance to let you chain method calls
     */
    header(...headers: [
        Header,
        ...Header[]
    ]): Fetch;
    /**
     * Use this method to set a custom timeout, instead of relying on brower default
     * timeouts like Chrome's 300 seconds default.
     *
     * @returns Returns the current instance to let you chain method calls
     */
    timeoutAfter(timeoutInMilliseconds: number): Fetch;
    /* All methods for setting body starts with the word 'body' */
    /**
     * Set the request body to be sent to server for HTTP methods such as POST/PUT.
     *
     * ### When to use this method?
     * For most users who want to send JSON data to the server, see the `bodyJSON` method
     * instead of a simpler API. For other types of data like `FormData`, `Blob` and `streams`
     * can just pass it into this method as the `body` parameter and the content type will be
     * automatically detected / set by the `fetch` function.
     *
     * ### Why have both `body` and `bodyJSON` method?
     * The reason for this method instead of just having `bodyJSON` only is because the library
     * cannot always just assume that users only use JSON data, and have to support data types
     * like FormData, Blob and etc... However the problem is that when content-type cannot be
     * set fetch will try to guess it, but when body is the results of JSON.stringify(this.#body),
     * fetch will guess the content-type to be text/plain and the browser will treat it as a safe
     * CORS request, which means that for that request there will be no pre-flight request sent.
     * Which means that the browser prevents certain headers from being used, which might cause
     * an issue, and also the server may not always respond correctly because they assume they got
     * text/plain even though your API endpoint is for application/json.
     *
     * The type of `body` value can be anything, as you can pass in any value that the
     * `fetch` API's `RequestInit`'s body property accepts.
     *
     * ### Using generics for TS Type Safety
     * ```javascript
     * const { res, err } = await oof
     *   .POST("/api")
     *   .body<FormData>(someValue) // TS will enforce that someValue must be FormData
     *   .run();
     * ```
     * The above code sets the body type for type safety.
     *
     * For TS users, the body type is a generic type variable even though its default type
     * for it is `any` so that you can use it to restrict the type passed into the method.
     * This allows you to enforce type safety where once a generic type is set, you know that
     * the value passed in for the `body` parameter cannot be any other type.
     *
     * ### On `optionalContentType`'s type safety
     * Note on `optionalContentType`'s type: Although it is possible to create a union type
     * of all allowed string literals for the content-type header / mime types, it is not
     * very feasible as it is a very big list that will be updated in the future. Therefore
     * users are expected to make sure that any string they pass is valid.
     * References on all supported content-type values:
     * - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type
     * - https://stackoverflow.com/questions/23714383/what-are-all-the-possible-values-for-http-content-type-header/48704300#48704300
     * - https://www.iana.org/assignments/media-types/media-types.xhtml
     *
     * @returns Returns the current instance to let you chain method calls
     */
    body<T = any>(body: T, optionalContentType?: string): Fetch;
    /**
     * @param data Any data type that is of 'application/json' type and can be stringified by JSON.stringify
     *
     * ### About
     * Method that stringifies a JSON stringifiable data type to use as the request body,
     * and sets the content-type to 'application/json'.
     *
     * ### What data type can be passed in?
     * The type of data that can be passed in, is any JS value that can be JSON serialized
     * with `JSON.stringify()`. See this [MDN link](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/JSON/stringify#description)
     * on what type of data can be passed in.
     *
     * ### What if you have no data to send?
     * Even though there is no default arguement, you do not have to call the `.data` method with an empty object when
     * calling a method like `oof.POST` as `fetch` and API services will just treat it as an empty object by default.
     *
     * ### Why does this method exists?
     * Because, the `#run` method needs to accept too many data input types,
     * so instead of doing all the processing and transforming then, the specific
     * transformations should be done in helper methods like this so that in the
     * actual `fetch` call, we can just set `body: this.#body` directly.
     *
     * ### Why not just use the `body` method?
     * Since JSON is one of the most popular methods of communication over HTTP,
     * this method helps users to write less code by helping them stringify their
     * JS object and set the content-type header to 'application/json', instead
     * of requiring them to explicitly call `.body(JSON.stringify(data), "application/json")`
     * every single time they want to send JSON data to their API servers.
     *
     * ### Using generics for TS Type Safety
     * ```javascript
     * const { res, err } = await oof
     *   .POST("/api")
     *   .bodyJSON<MyRequestBody>(val) // TS will enforce that val must be MyRequestBody
     *   .run();
     * ```
     * The above code sets the body type for type safety.
     *
     * For TS users, the data parameter have a generic type even though its default type
     * is `any` so that you can use it to restrict the type passed into the method.
     * This allows you to enforce type safety where once a generic type is set, you know that
     * the value passed in for the `body` parameter cannot be any other type.
     *
     * @returns Returns the current instance to let you chain method calls
     */
    bodyJSON<T = any>(data: T): Fetch;
    /**
     * # Warning
     * This method is generally not used since this returns the raw HTTP Response object.
     * Library users should use the other run methods that handle value extraction too,
     * such as `runJSON`, `runFormData` and etc... This method is made available as an
     * escape hatch for users who do not want value extraction done, so that they can build
     * extra logic on top of this library by getting the raw Response object back.
     *
     * Safe version of the `#run` method that **will not throw** or let any errors bubble up,
     * i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.useDefault().GET("/api").run();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be the Response object
     * ```
     */
    run(): Promise<{
        res: undefined;
        err: RequestError;
    } | {
        res: Response;
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `#run` method to return response body parsed as text.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.useDefault().GET("/api").runText();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be be a string
     * ```
     */
    runText(optionalValidator?: Validator<string>): Promise<{
        res: undefined;
        err: RequestError;
    } | {
        res: {
            ok: boolean;
            status: number;
            headers: Headers;
            data: string;
        };
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `#run` method to return response body parsed as Blob.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.useDefault().GET("/api").runBlob();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be a Blob
     * ```
     */
    runBlob(optionalValidator?: Validator<Blob>): Promise<{
        res: undefined;
        err: RequestError;
    } | {
        res: {
            ok: boolean;
            status: number;
            headers: Headers;
            data: Blob;
        };
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `#run` method to return response body parsed as FormData.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.useDefault().GET("/api").runFormData();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be form data
     * ```
     */
    runFormData(optionalValidator?: Validator<FormData>): Promise<{
        res: undefined;
        err: RequestError;
    } | {
        res: {
            ok: boolean;
            status: number;
            headers: Headers;
            data: FormData;
        };
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `#run` method to return response body parsed as ArrayBuffer.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.useDefault().GET("/api").runArrayBuffer();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be an Array Buffer
     * ```
     */
    runArrayBuffer(optionalValidator?: Validator<ArrayBuffer>): Promise<{
        res: undefined;
        err: RequestError;
    } | {
        res: {
            ok: boolean;
            status: number;
            headers: Headers;
            data: ArrayBuffer;
        };
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `#run` method to return response body parsed as JSON.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.useDefault().GET("/api").runJSON<MyResponseObjectType>();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be 'MyResponseObjectType'
     * ```
     *
     * ### Using generics for TS Type Safety
     * For TS users, this method accepts a generic type to type the returned object. Allowing you to have type
     * safety for the response object. However, this DOES NOT perform any runtime data validation, so even if
     * it is type safe, it does not mean that the response object is guaranteed to be what you typed it to be.
     *
     * For run time type safety, please pass in a type predicate validator to do response data validation before
     * type narrowing it down to the generic type T passed in.
     */
    runJSON<T = JsonResponse>(optionalValidator?: Validator<T>): Promise<{
        res: undefined;
        err: RequestError;
    } | {
        res: {
            ok: boolean;
            status: number;
            headers: Headers;
            data: T;
        };
        err: undefined;
    }>;
}
/**
 * Class used to implement the Builder pattern for creating `Fetch` instances.
 * This class should only be used internally by `oof` and should not be exposed
 * to library users since they should not be using this to make API calls.
 */
declare class Builder {
    #private;
    constructor(baseUrl: string);
    /**
     * Method to set default options that will be applied to all API calls made
     * through the `Fetch` instance created by this `Builder` instance.
     * The options set here can be overwritten one-off in specific API calls using
     * the `options` method of `Fetch` class.
     *
     * Note that everytime this is called, all the default options are overwritten
     * to use the options passed in as the new default options without merging with
     * the original default options.
     *
     * @returns Returns the current instance to let you chain method calls
     */
    setDefaultOptions(opts: RequestInit): Builder;
    /**
     * Method to set default headers that will be applied to all API calls made
     * through the `Fetch` instance created by this `Builder` instance.
     * The headers set here can be overwritten one-off in specific API calls using
     * the `header` method of `Fetch` class.
     *
     * Useful for doing things like passing in a function to generate auth header,
     * so that the function does not need to be passed in for every API call made
     * with the `Fetch` instance. Another example is if every API call needs a
     * recaptcha token, instead of passing in the recaptcha token header generator
     * function everywhere, this can be set just once as a default header value.
     *
     * Note that everytime this is called, the default headers is fully overwritten
     * to use the headers passed in as the new default headers without merging with
     * the original default headers.
     *
     * The `headers` function parameter forces the caller to pass in at least one
     * argument for this variadic method. See https://stackoverflow.com/a/72286990
     *
     * @returns Returns the current instance to let you chain method calls
     */
    setDefaultHeaders(...headers: [
        Header,
        ...Header[]
    ]): Builder;
    /**
     * tl;dr Do not use this unless you know what you are doing.
     *
     * ### About
     * Wrapper function over `Fetch` constructor to construct a new
     * instance with the given HTTP method and path alongside the
     * values set on this `Builder` instance.
     *
     * ### Library Users
     * For the most part, library users ***SHOULD NOT*** be using this
     * low level API, instead they should use the methods named after
     * the HTTP methods such as builder.GET, builder.POST and so on.
     * The only reason why this is made public is for users who want
     * to use HTTP methods like `HEAD` and `OPTION` to still be able
     * to do so using our library instead of doing some crazy hack,
     * since there is no method for those specific HTTP methods.
     */
    HTTP: (method: HTTPMethod, path: string) => Fetch;
    /** Construct a new `Fetch` instance to make a `GET` API call */
    GET: (path?: string) => Fetch;
    /** Construct a new `Fetch` instance to make a `POST` API call */
    POST: (path?: string) => Fetch;
    /** Construct a new `Fetch` instance to make a `PUT` API call */
    PUT: (path?: string) => Fetch;
    /** Construct a new `Fetch` instance to make a `PATCH` API call */
    PATCH: (path?: string) => Fetch;
    /** Construct a new `Fetch` instance to make a `DELETE` API call */
    DEL: (path?: string) => Fetch;
}
/**
 * Class used to create an Object Oriented `Fetch` abstraction.
 * This uses the object oriented builder pattern to give users
 * an easy to use chainable interface to build their API calls.
 */
declare class oof {
    #private;
    /**
     * Add a new baseUrl identifier and baseUrl mapping.
     * For example, you can add different baseUrls for your APIs such
     * as 'v1', 'v2', 'latest', 'billing' and etc...
     */
    static addBase(identifier: string, url: string): typeof oof;
    /**
     * Select a baseUrl to use, using a baseUrl identifier.
     * For example, you can select the 'v1' baseUrl after adding it
     * using the `addBase` static method.
     */
    static useBase(identifier: string): Builder;
    /**
     * Set a particular identifier and baseUrl mapping as the default
     * mapping to use, so that you can easily use this default mapping in
     * your API calls without having to import and specify the identifier.
     */
    static setDefault(identifier: string): void;
    /**
     * Use the default identifier and baseUrl mapping for an API call.
     */
    static useDefault(): Builder;
    /**
     * Use a Full URL string (with API path) to make an API call without using any baseUrl.
     *
     * This method should be used when making one off API calls, without having to set
     * a `baseUrl` first. Usually used when you need to make an API call to another domain.
     * E.g. calling a third party API for some integration.
     */
    static useOnce(fullUrlString: string): Builder;
}
/**
 * Utility to convert a zod parser to validator function for validating
 * response data in `Fetch` run methods.
 *
 * Alternative conversion using the zod `safeParse` method instead.
 * The problem with this, is that no errors are thrown from the parser,
 * which means that the API lib will throw its own Error instance with
 * a Validation Failed error message, while it might be preferrable for
 * users to receive the actual `ZodError` by having it bubble through
 * our API library. So that users can get it with `const { res, err } =`
 * and they can check `err` type using `instanceof`.
 * ```typescript
 * const zodToValidator =
 *   <ZodParserType extends ZodType>(parser: ZodParserType) =>
 *   (data: unknown): data is z.infer<ZodParserType> =>
 *     parser.safeParse(data).success;
 * ```
 */
declare const zodToValidator: <ZodParserType extends z.ZodType<any, z.ZodTypeDef, any>>(parser: ZodParserType) => Validator<z.TypeOf<ZodParserType>>;
/**
 * Barrel file to export everything a library user can access.
 */
export { oof, zodToValidator };
export type { HTTPMethod, Header, HeaderValue, RequestError, ApiResponse, Validator };
